#!/usr/bin/python2
# $Id: check_rsnap.py,v 1.3 2013/03/31 16:46:46 root Exp $
#  check_rsnap.py
#
#  Copyright 2013 Rich Braun <richb@instantlinux.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import sys
import time
import argparse
import MySQLdb

def argument_parser():
    parser = argparse.ArgumentParser(description='Check db for rsnap backups')
    parser.add_argument('-H',
        help = 'host to query (Default: all)',
        dest = 'host',
        type = str,
        default = 'all'
        )
    parser.add_argument('-w',
        help = 'WARNING if number of backups too low (Default: 1)',
        dest = 'warning',
        type = int,
        default = 1
        )
    parser.add_argument('-c',
        help = 'CRITICAL if number of backups too low (Default: 1)',
        dest = 'critical',
        type = int,
        default = 1
        )
    parser.add_argument('-f',
        help = 'WARNING if number of files too low (Default: 1000)',
        dest = 'file_warning',
        type = int,
        default = 1000
        )
    parser.add_argument('-i',
        help = 'backup interval, in hours (Default: 24)',
        dest = 'interval',
        type = int,
        default = '24'
        )
    parser.add_argument('-d',
        help = 'database host (Default: localhost)',
        dest = 'dbhost',
        type = str,
        default = 'localhost'
        )
    parser.add_argument('-P',
        help = 'database port (Default: 3306)',
        dest = 'dbport',
        type = int,
        default = '3306'
        )
    parser.add_argument('-n',
        help = 'name of database (Default: rsnap)',
        dest = 'dbname',
        type = str,
        default = 'rsnap'
        )
    parser.add_argument('-u',
        help = 'database user (Default: nagmon)',
        dest = 'dbuser',
        type = str,
        default = 'nagmon'
        )
    parser.add_argument('-p',
		    help = 'database password',
        dest = 'dbpass',
        type = str
        )
    return parser.parse_args()

def main():
    args = argument_parser()
    start_time = time.time()
    db = MySQLdb.connect(host=args.dbhost, port=args.dbport, user=args.dbuser, passwd=args.dbpass, db=args.dbname)
    cur = db.cursor()

    # Look for recent backups
    q = 'SELECT COUNT(*) FROM savesets WHERE finished > (NOW() - INTERVAL ' + str(args.interval) + ' HOUR)'
    if args.host != 'all':
        q = q + ' AND host_id=(SELECT id FROM hosts WHERE hostname="' + args.host + '" LIMIT 1)'
    cur.execute(q)
    row = cur.fetchone()
    num_backups = row[0]

    # Count files in backups
    q = 'SELECT files FROM savesets JOIN hosts on hosts.id=savesets.host_id \
         WHERE finished > (now() - INTERVAL ' + str(args.interval) + ' HOUR)'
    if args.host != 'all':
        q = q + ' AND hostname="' + args.host + '"'
    cur.execute(q)
    row = cur.fetchone()
    num_files = row[0]

    rsnap_stat = 'OK: rsnap'
    exit_value = 0
    if num_backups < args.critical:
        rsnap_stat = 'CRITICAL: rsnap'
	exit_value = 2
    elif num_backups < args.warning:
        rsnap_stat = 'WARNING: rsnap'
	exit_value = 1
    elif num_files < args.file_warning:
        rsnap_stat = 'WARNING: rsnap'
        exit_value = 1
    sys.stdout.write(rsnap_stat + " backups:" + str(num_backups) + " files:" + str(num_files) + "\n")
    sys.exit((exit_value))

if __name__ == '__main__':
    main()
    sys.stdout.write("Something really bad happened!")
    sys.exit((3))



"""models

Database model definitions for SQLalchemy

created 28-jul-2018 by richb@instantlinux.net

license: lgpl-2.1
"""

# coding: utf-8
from sqlalchemy import BIGINT, BOOLEAN, Column, Enum, Float, ForeignKey, \
     INTEGER, Index, String, TIMESTAMP, text, VARBINARY
from sqlalchemy import func
from sqlalchemy.dialects import sqlite
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata
BigIntId = BIGINT().with_variant(sqlite.INTEGER(), 'sqlite')


class ConfigTable(Base):
    __tablename__ = 'config'
    __table_args__ = (
        Index('index1', 'keyword', 'host_id', unique=True),
    )

    id = Column(INTEGER, primary_key=True, unique=True,
                autoincrement=True)
    keyword = Column(String(32), nullable=False)
    value = Column(String(1023), nullable=True)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    host_id = Column(ForeignKey(u'hosts.id'), nullable=False, index=True)

    host = relationship('Host')


class Host(Base):
    __tablename__ = 'hosts'

    id = Column(INTEGER, primary_key=True, unique=True,
                autoincrement=True)
    hostname = Column(String(45), nullable=False, unique=True)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())


class File(Base):
    __tablename__ = 'files'
    __table_args__ = (
        Index('index3', 'filename', 'path', 'host_id', 'mode', 'size', 'mtime',
              'uid', 'gid', unique=True),
    )

    id = Column(BigIntId, primary_key=True, nullable=False, unique=True,
                autoincrement=True)
    path = Column(String(1023), nullable=False)
    filename = Column(String(255), nullable=False)
    owner = Column(String(48))
    grp = Column(String(48))
    uid = Column(INTEGER, nullable=False)
    gid = Column(INTEGER, nullable=False)
    mode = Column(INTEGER, nullable=False)
    size = Column(BIGINT, nullable=False)
    ctime = Column(TIMESTAMP)
    mtime = Column(TIMESTAMP)
    type = Column(Enum(u'c', u'd', u'f', u'l', u's'), nullable=False)
    links = Column(INTEGER, nullable=False, server_default=text("1"))
    sparseness = Column(Float, nullable=False, server_default=text("1"))
    shasum = Column(VARBINARY(64))
    first_backup = Column(TIMESTAMP, nullable=False, server_default=func.now())
    last_backup = Column(TIMESTAMP)
    # host_id = Column(ForeignKey(u'hosts.id'), primary_key=True,
    #                 nullable=False, index=True)
    host_id = Column(ForeignKey(u'hosts.id'), nullable=False, index=True)

    host = relationship('Host')


class Saveset(Base):
    __tablename__ = 'savesets'

    id = Column(INTEGER, primary_key=True, nullable=False, unique=True,
                autoincrement=True)
    saveset = Column(String(45), nullable=False, unique=True)
    location = Column(String(32))
    files = Column(BIGINT)
    size = Column(BIGINT)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    finished = Column(TIMESTAMP, index=True)
    # host_id = Column(ForeignKey(u'hosts.id'), primary_key=True,
    #                  nullable=False, index=True)
    # backup_host_id = Column(ForeignKey(u'hosts.id'), primary_key=True,
    #                         nullable=False, index=True)
    host_id = Column(ForeignKey(u'hosts.id'), nullable=False, index=True)
    backup_host_id = Column(ForeignKey(u'hosts.id'), nullable=False,
                            index=True)

    backup_host = relationship(
        'Host', primaryjoin='Saveset.backup_host_id == Host.id')
    host = relationship('Host', primaryjoin='Saveset.host_id == Host.id')


class Volume(Base):
    __tablename__ = 'volumes'

    id = Column(INTEGER, primary_key=True, nullable=False, unique=True,
                autoincrement=True)
    volume = Column(String(45), nullable=False, unique=True)
    path = Column(String(255), nullable=False)
    size = Column(BIGINT)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    removable = Column(BOOLEAN, nullable=False, server_default=text("0"))
    mounted = Column(BOOLEAN, nullable=False, server_default=text("1"))
    # host_id = Column(ForeignKey(u'hosts.id'), primary_key=True,
    #                  nullable=False, index=True, server_default=text("0"))
    host_id = Column(ForeignKey(u'hosts.id'), nullable=False,
                     index=True, server_default=text("0"))

    host = relationship('Host')


class AlembicVersion(Base):
    __tablename__ = 'alembic_version'

    version_num = Column(String(32), primary_key=True, nullable=False)



"""Initialize database

Revision ID: ab8ec674d281
Revises:
Create Date: 2018-08-13 08:26:26.508945

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'ab8ec674d281'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'hosts',
        sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column('hostname', sa.String(length=45), nullable=False),
        sa.Column('created', sa.TIMESTAMP(), server_default=sa.func.now(),
                  nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('hostname'),
        sa.UniqueConstraint('id')
    )
    op.create_table(
        'files',
        sa.Column('id', sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column('path', sa.String(length=1023), nullable=False),
        sa.Column('filename', sa.String(length=255), nullable=False),
        sa.Column('owner', sa.String(length=48), nullable=True),
        sa.Column('grp', sa.String(length=48), nullable=True),
        sa.Column('uid', sa.INTEGER(), nullable=False),
        sa.Column('gid', sa.INTEGER(), nullable=False),
        sa.Column('mode', sa.INTEGER(), nullable=False),
        sa.Column('size', sa.BIGINT(), nullable=False),
        sa.Column('ctime', sa.TIMESTAMP(), nullable=True),
        sa.Column('mtime', sa.TIMESTAMP(), nullable=True),
        sa.Column('type', sa.Enum(u'c', u'd', u'f', u'l', u's'),
                  nullable=False),
        sa.Column('links', sa.INTEGER(), server_default=sa.text(u'1'),
                  nullable=False),
        sa.Column('sparseness', sa.Float(), server_default=sa.text(u'1'),
                  nullable=False),
        sa.Column('shasum', sa.VARBINARY(length=64), nullable=True),
        sa.Column('first_backup', sa.TIMESTAMP(), server_default=sa.func.now(),
                  nullable=False),
        sa.Column('last_backup', sa.TIMESTAMP(), nullable=True),
        sa.Column('host_id', sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(['host_id'], [u'hosts.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id')
    )
    op.create_index('index3', 'files', [
        'filename', 'path', 'host_id', 'mode', 'size', 'mtime', 'uid', 'gid'],
        unique=True)
    op.create_index(op.f('ix_files_host_id'), 'files', ['host_id'],
                    unique=False)
    op.create_table(
        'savesets',
        sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column('saveset', sa.String(length=45), nullable=False),
        sa.Column('location', sa.String(length=32), nullable=True),
        sa.Column('created', sa.TIMESTAMP(), server_default=sa.func.now(),
                  nullable=False),
        sa.Column('finished', sa.TIMESTAMP(), nullable=True),
        sa.Column('host_id', sa.INTEGER(), nullable=False),
        sa.Column('backup_host_id', sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(['backup_host_id'], [u'hosts.id'], ),
        sa.ForeignKeyConstraint(['host_id'], [u'hosts.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id'),
        sa.UniqueConstraint('saveset')
    )
    op.create_index(op.f('ix_savesets_backup_host_id'), 'savesets',
                    ['backup_host_id'], unique=False)
    op.create_index(op.f('ix_savesets_finished'), 'savesets', ['finished'],
                    unique=False)
    op.create_index(op.f('ix_savesets_host_id'), 'savesets', ['host_id'],
                    unique=False)
    op.create_table(
        'volumes',
        sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column('volume', sa.String(length=45), nullable=False),
        sa.Column('path', sa.String(length=255), nullable=False),
        sa.Column('size', sa.BIGINT(), nullable=True),
        sa.Column('created', sa.TIMESTAMP(), server_default=sa.func.now(),
                  nullable=False),
        sa.Column('removable', sa.BOOLEAN(), server_default=sa.text(u'0'),
                  nullable=False),
        sa.Column('mounted', sa.BOOLEAN(), server_default=sa.text(u'1'),
                  nullable=False),
        sa.Column('host_id', sa.INTEGER(), server_default=sa.text(u'0'),
                  nullable=False),
        sa.ForeignKeyConstraint(['host_id'], [u'hosts.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id'),
        sa.UniqueConstraint('volume')
    )
    op.create_index(op.f('ix_volumes_host_id'), 'volumes', ['host_id'],
                    unique=False)
    op.create_table(
        'backups',
        sa.Column('id', sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column('saveset_id', sa.INTEGER(), nullable=False),
        sa.Column('volume_id', sa.INTEGER(), nullable=False),
        sa.Column('file_id', sa.BIGINT(), nullable=False),
        sa.ForeignKeyConstraint(['file_id'], [u'files.id'], ),
        sa.ForeignKeyConstraint(['saveset_id'], [u'savesets.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['volume_id'], [u'volumes.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id')
    )
    op.create_index(op.f('ix_backups_file_id'), 'backups', ['file_id'],
                    unique=False)
    op.create_index(op.f('ix_backups_saveset_id'), 'backups', ['saveset_id'],
                    unique=False)
    op.create_index(op.f('ix_backups_volume_id'), 'backups', ['volume_id'],
                    unique=False)
    # ### end Alembic commands ###


"""
def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_backups_volume_id'), table_name='backups')
    op.drop_index(op.f('ix_backups_saveset_id'), table_name='backups')
    op.drop_index(op.f('ix_backups_file_id'), table_name='backups')
    op.drop_table('backups')
    op.drop_index(op.f('ix_volumes_host_id'), table_name='volumes')
    op.drop_table('volumes')
    op.drop_index(op.f('ix_savesets_host_id'), table_name='savesets')
    op.drop_index(op.f('ix_savesets_finished'), table_name='savesets')
    op.drop_index(op.f('ix_savesets_backup_host_id'), table_name='savesets')
    op.drop_table('savesets')
    op.drop_index(op.f('ix_files_host_id'), table_name='files')
    op.drop_index('index3', table_name='files')
    op.drop_table('files')
    op.drop_table('hosts')
    # ### end Alembic commands ###
"""



#!/usr/bin/env python3
"""Linux-based backup utility with integrity-verification

created 28-jul-2018 by richb@instantlinux.net

Usage:
  secondshot [--action=ACTION] [--dbhost=HOST] [--dbuser=USER] [--dbpass=PASS]
           [--dbname=DB] [--dbport=PORT] [--dbtype=TYPE] [--db-url=URL]
           [--backup-host=HOST] [--host=HOST]... [--logfile=FILE]
           [--list-hosts] [--list-savesets] [--list-volumes]
           [--filter=STR] [--format=FORMAT] [--hashtype=ALGORITHM]
           [--manifest=FILE] [--rsnapshot-conf=FILE] [--autoverify=BOOL]
           [--sequence=VALUES] [--volume=VOL] [--log-level=STR]
           [--version] [-v]...
  secondshot --action=start --host=HOST --volume=VOL [--autoverify=BOOL]
           [--log-level=STR] [-v]...
  secondshot --action=rotate --interval=INTERVAL [--logfile=FILE]
           [--log-level=STR] [--rsnapshot-conf=FILE] [-v]...
  secondshot --verify=SAVESET... [--format=FORMAT] [--hashtype=ALG]
           [--logfile=FILE] [--log-level=STR] [--rsnapshot-conf=FILE] [-v]...
  secondshot --action=schema-update [-v]...
  secondshot (-h | --help)

Options:
  --action=ACTION       Action to take (archive, rotate, start)
  --backup-host=HOST    Hostname taking the backup (default hostname -s)
  --dbhost=HOST         DB host (default: db00)
  --dbname=DB           DB name (default: secondshot)
  --dbport=PORT         DB port (default: 3306)
  --dbuser=USER         DB user (default: bkp)
  --dbpass=PASS         DB password (default env variable DBPASS)
  --dbtype=TYPE         DB type, e.g. mysql+pymysql (default: sqlite)
  --db-url=URL          Full URL (alternative to above DB specifiers)
  --host=HOST           Source host(s) to back up
  --interval=INTERVAL   Rotation interval: e.g. hourly, daysago
  --list-hosts          List hosts
  --list-savesets       List savesets
  --list-volumes        List volumes
  --filter=STR          Filter to limit listing [default: *]
  --format=FORMAT       Format (text or json) [default: text]
  --logfile=FILE        Logging destination [default: /var/log/secondshot]
  --log-level=STR       Syslog level debug/info/warn/none [default: info]
  --manifest=FILE       Name of manifest file [default: .secondshot-manifest]
  --rsnapshot-conf=FILE Path of rsnapshot's config file
                        (default: /etc/backup-daily.conf)
  --sequence=VALUES     Sequence of retention intervals
                        [default: hourly,daysago,weeksago,monthsago,\
semiannually,yearsago]
  --autoverify=BOOL     Verify each just-created saveset (default: yes)
  --hashtype=ALGORITHM  Hash algorithm md5, sha256, sha512 (default: md5)
  --verify=SAVESET      Verify checksums of stored files
  --version             Display software version
  --volume=VOLUME       Volume for storing saveset
  -v --verbose          Verbose output
  -h --help             List options

license: lgpl-2.1
"""

import docopt
import json
import sys

from secondshot.actions import Actions
from secondshot.config import Config
from secondshot.syslogger import Syslog
from secondshot._version import __version__


def main():
    opts = Config().docopt_convert(docopt.docopt(__doc__))
    Syslog.logger = Syslog(opts)
    obj = Actions(opts)

    result = {}
    status = 'ok'

    if (opts['list-hosts']):
        result = obj.list_hosts()
    elif (opts['list-savesets']):
        result = obj.list_savesets()
    elif (opts['list-volumes']):
        result = obj.list_volumes()
    elif (opts['verify']):
        result = obj.verify(opts['verify'])
    elif (opts['version']):
        result = dict(version=[dict(name='secondshot %s' % __version__)])
    elif (opts['action'] == 'start'):
        result = obj.start(obj.hosts, obj.volume)
        status = result['start']['status']
    elif (opts['action'] == 'rotate'):
        result = obj.rotate(opts['interval'])
    elif (opts['action'] == 'schema-update'):
        result = obj.schema_update()
        status = result['status']
    else:
        sys.exit('Unknown action: %s' % opts['action'])

    if (opts['format'] == 'json'):
        sys.stdout.write(json.dumps(result) + '\n')
    elif (opts['format'] == 'text' and result and next(iter(result.keys())) in
          ['hosts', 'savesets', 'schema-update', 'version', 'volumes']):
        for item in result[next(iter(result.keys()))]:
            sys.stdout.write(item['name'] + '\n')
    if (status != 'ok'):
        exit(1)


if __name__ == '__main__':
    main()



"""constants

Constant-value definitions for secondshot

created 17-aug-2018 by richb@instantlinux.net

license: lgpl-2.1
"""


class Constants(object):
    DBPASS_FILE = '/run/secrets/secondshot-db-password'
    DBFILE_PATH = '/metadata'
    DBOPTS_ALLOW = ['autoverify', 'hashtype', 'host', 'rsnapshot-conf',
                    'volume']
    DEFAULT_VOLUME = 'backup'
    MAX_INSERT = 2000
    OPTS_DEFAULTS = {
        'autoverify': 'yes',
        'dbhost': 'db00',
        'dbname': 'secondshot',
        'dbpass': None,
        'dbport': '3306',
        'dbtype': 'sqlite',
        'dbuser': 'bkp',
        'db-url': None,
        'hashtype': 'md5',
        'manifest': '.snapshot-manifest',
        'rsnapshot-conf': '/etc/backup-daily.conf'}
    SNAPSHOT_ROOT = '/backups'
    SYNC_PATH = '.sync'



"""config

Configuration-file parsing

created 11-aug-2018 by richb@instantlinux.net

license: lgpl-2.1
"""

import os
import sqlalchemy.exc
import sys

from secondshot.constants import Constants
from secondshot.models import ConfigTable, Host
from secondshot.syslogger import Syslog


class Config(object):

    autoverify = Constants.OPTS_DEFAULTS['autoverify']
    hashtype = Constants.OPTS_DEFAULTS['hashtype']
    manifest = Constants.OPTS_DEFAULTS['manifest']
    rsnapshot_conf = Constants.OPTS_DEFAULTS['rsnapshot-conf']
    sequence = None
    snapshot_root = Constants.SNAPSHOT_ROOT

    def init_db_get_config(self, db_session, hostname):
        """Initialize db session and read host-specific entries from
        config table

        Args:
            db_session (obj): sqlalchemy session
            hostname (str): fetch only these host-specific settings
        Returns:
            result (dict):  key/value pairs found in database
        """

        self.session = db_session
        self.hostname = hostname
        result = {}
        for record in self.session.query(ConfigTable).join(
                ConfigTable.host).filter(Host.hostname == hostname):
            if (record.keyword == 'host'):
                result[record.keyword] = record.value.split(',')
            else:
                result[record.keyword] = record.value
        return result

    def db_set(self, keyword, value):
        """Update a config setting

        Args:
            keyword: Config key
            value: New value
        """
        try:
            record = self.session.query(ConfigTable).join(
                ConfigTable.host).filter(
                    ConfigTable.keyword == keyword,
                    Host.hostname == self.hostname).one()
            record.value = value
        except sqlalchemy.orm.exc.NoResultFound:
            host = self.session.query(Host).filter(
                Host.hostname == self.hostname).one()
            record = ConfigTable(keyword=keyword, value=value,
                                 host=host)
        self.session.add(record)
        self.session.commit()

    def db_get(self, keyword):
        """Fetch a config setting

        Args:
            keyword: Config key
        """
        try:
            record = self.session.query(ConfigTable).join(
                ConfigTable.host).filter(
                    ConfigTable.keyword == keyword,
                    Host.hostname == self.hostname).one()
            return record.value
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    @staticmethod
    def get_db_url(opts):
        """Get the DB url from command-line options

        Args:
            opts (dict): options dictionary
        Returns:
            result (str): URL of database
        Raises:
            RuntimeError
        """

        for item in ['dbhost', 'dbname', 'dbpass', 'dbport', 'dbtype',
                     'dbuser', 'db-url']:
            if (not opts[item]):
                opts[item] = os.environ.get(item.upper().replace('-', '_'),
                                            Constants.OPTS_DEFAULTS[item])
        if (opts['db-url']):
            return opts['db-url']
        else:
            if (opts['dbtype'] == 'sqlite'):
                return ('sqlite:///%(path)s/%(database)s' % {
                    'path': Constants.DBFILE_PATH,
                    'database': opts['dbname']
                })
            else:
                pw = opts['dbpass']
                if (not pw and os.path.isfile(Constants.DBPASS_FILE)):
                    pw = open(Constants.DBPASS_FILE, 'r').read()
                if (not pw):
                    raise RuntimeError('Database password is not set')
                return ('%(dbtype)s://%(user)s:%(password)s@'
                        '%(endpoint)s/%(database)s' % {
                            'dbtype': opts['dbtype'],
                            'user': opts['dbuser'],
                            'password': pw,
                            'endpoint': '%s:%d' % (opts['dbhost'],
                                                   int(opts['dbport'])),
                            'database': opts['dbname']
                        })

    def set_opts(self, cli_opts, db_session, hostname):
        """Read additional options from ConfigTable in database, apply
        precedence rules (there are limitations due to issue cited at
        https://github.com/docopt/docopt/issues/341).

        Args:
            cli_opts (dict): options set on CLI by docopt
            db_session (object): database session
            hostname (str): hostname key for config table lookup
        Returns:
            result (dict): options values after order of precedence
        Raises:
            ValueError
        """

        try:
            opts = self.init_db_get_config(db_session, hostname)
            self.validate_configs(opts, Constants.DBOPTS_ALLOW)
        except ValueError as ex:
            sys.exit('ERROR: DB configuration %s' % str(ex))
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            Syslog.logger.warn('DB config message=%s' % str(ex))
            opts = {}
        for key, value in cli_opts.items():
            if (key not in opts):
                opts[key] = value
        for key, value in Constants.OPTS_DEFAULTS.items():
            if (opts[key] is None):
                opts[key] = value
        try:
            self.validate_configs(opts, cli_opts.keys())
        except ValueError as ex:
            sys.exit('ERROR: %s' % str(ex))

        if (opts['autoverify'].lower() in ['false', 'no', 'off']):
            Config.autoverify = False
        elif (opts['autoverify'].lower() in ['true', 'yes', 'on']):
            Config.autoverify = True
        Config.hashtype = opts['hashtype']
        Config.manifest = opts['manifest']
        Config.rsnapshot_conf = opts['rsnapshot-conf']
        Config.sequence = opts['sequence'].split(',')
        Config.snapshot_root = Constants.SNAPSHOT_ROOT
        return opts

    def validate_configs(self, opts, valid_choices):
        """Validate configuration settings; jsonschema would be
        tidier--but overkill.

        Args:
            opts (dict): options dictionary
            valid_choices (list): list of valid keywords
        Raises:
            ValueError if any setting's keyword or value is incorrect
        """

        for keyword, value in opts.items():
            if (keyword == 'format'):
                if (value not in ['json', 'text']):
                    raise ValueError(
                        'format=%s not json or text' % value)
            elif (keyword == 'hashtype'):
                if (value not in ['md5', 'sha256', 'sha512']):
                    raise ValueError(
                        'hashtype=%s not md5, sha256 or sha512' % value)
            elif (keyword == 'autoverify'):
                if (value not in ['false', 'no', 'off', 'true', 'yes', 'on']):
                    raise ValueError(
                        'autoverify=%s invalid boolean value' % value)

            if (keyword not in valid_choices):
                raise ValueError(
                    'keyword=%s unrecognized option setting' % keyword)

    @staticmethod
    def docopt_convert(opts):
        """Convert from dash-dash keyword to just keyword

        Args:
            opts (dict): options dictionary
        Returns:
            result (dict):  command-line options dict, with updated keys
        """

        return {key.strip('-'): value for (key, value) in opts.items()}

    def rsnapshot_cfg(self):
        """Parse the rsnapshot config file into a dictionary
        Keywords in this config file can have up to two parameters;
        for those which allow multiple statements of the same keyword,
        return a 2-level sub-dictionary or a single-level list

        Returns:
            dict:  parsed contents
        Raises:
            SyntaxError: if unexpected syntax
        """

        # Keywords that can have multiple settings in rsnapshot.conf
        self.rsnap_multiple = ['backup', 'backup_script', 'exclude',
                               'include', 'include_conf', 'interval',
                               'retain']
        filename = Config.rsnapshot_conf

        contents = {}
        try:
            fp = open(filename, 'r')
        except IOError as ex:
            Syslog.logger.warn('Cannot read rsnapshot_conf=%s, message=%s'
                               % (filename, str(ex)))
            return contents
        linenum = 1
        for line in fp:
            if '#' in line:
                line, comment = line.split('#', 1)
            tokens = line.strip().split(None, 2)
            if (len(tokens) == 0):
                continue
            elif (len(tokens) < 2):
                raise SyntaxError('file=%s at line %d\n%s' % (
                    filename, linenum, line))
            key = tokens[0]
            if (key in self.rsnap_multiple):
                if (len(tokens) == 2):
                    if (key not in contents):
                        contents[key] = []
                    contents[key].append(tokens[1])
                else:
                    if (key not in contents):
                        contents[key] = {}
                    contents[key][tokens[1]] = tokens[2]
            elif (key not in contents):
                contents[key] = ' '.join(tokens[1:])
            else:
                raise SyntaxError('file=%s (%d): duplicate keyword %s' % (
                    filename, linenum, key))
            linenum += 1
        fp.close()
        return contents



"""drop manifests

Revision ID: 6cf114485e53
Revises: 23296991a8e5
Create Date: 2020-10-17 16:27:26.467159

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = '6cf114485e53'
down_revision = '23296991a8e5'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('backups')
    op.create_index('index1', 'config', ['keyword', 'host_id'], unique=True)
    op.create_index(op.f('ix_config_host_id'), 'config', ['host_id'], unique=False)
    with op.batch_alter_table('config', schema=None) as batch_op:
        batch_op.create_foreign_key('config_ibfk_1', 'hosts', ['host_id'], ['id'])

    with op.batch_alter_table('files', schema=None) as batch_op:
        batch_op.alter_column('path',
               existing_type=mysql.VARCHAR(length=512),
               nullable=False)
    op.add_column('savesets', sa.Column('files', sa.BIGINT(), nullable=True))
    op.add_column('savesets', sa.Column('size', sa.BIGINT(), nullable=True))

    # op.drop_index('fk_backups_files', table_name='backups')
    # op.drop_index('id_UNIQUE', table_name='backups')
    # op.create_index(op.f('ix_files_host_id'), 'files', ['host_id'], unique=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    """
    op.drop_index(op.f('ix_volumes_host_id'), table_name='volumes')
    op.drop_column('savesets', 'size')
    op.drop_column('savesets', 'files')
    op.drop_index(op.f('ix_files_host_id'), table_name='files')
    op.alter_column('files', 'path',
               existing_type=mysql.VARCHAR(length=512),
               nullable=True)
    op.drop_constraint(None, 'config', type_='foreignkey')
    op.drop_index(op.f('ix_config_host_id'), table_name='config')
    op.drop_index('index1', table_name='config')
    op.create_table('backups',
    sa.Column('id', mysql.BIGINT(display_width=20, unsigned=True), autoincrement=True, nullable=False),
    sa.Column('saveset_id', mysql.INTEGER(display_width=10, unsigned=True), autoincrement=False, nullable=False),
    sa.Column('volume_id', mysql.INTEGER(display_width=10, unsigned=True), autoincrement=False, nullable=False),
    sa.Column('file_id', mysql.BIGINT(display_width=20, unsigned=True), autoincrement=False, nullable=False),
    sa.ForeignKeyConstraint(['file_id'], ['files.id'], name='backups_ibfk_1'),
    sa.ForeignKeyConstraint(['saveset_id'], ['savesets.id'], name='fk_backups_savesets1', ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['volume_id'], ['volumes.id'], name='fk_backups_volumes1'),
    sa.PrimaryKeyConstraint('id'),
    mysql_default_charset='utf8',
    mysql_engine='InnoDB'
    )
    op.create_index('id_UNIQUE', 'backups', ['id'], unique=True)
    op.create_index('fk_backups_files', 'backups', ['file_id'], unique=False)
    """
    # ### end Alembic commands ###
    pass



import io
import os.path
import setuptools
from setuptools.command.test import test as TestCommand

from secondshot._version import __version__

__long_desc__ = io.open(os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'README.md'), encoding='utf-8').read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--junitxml', 'tests/test-result.xml',
            '--cov-report', 'term-missing',
            '--cov-report', 'html',
            '--cov-report', 'xml']
        self.test_suite = True

    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


setuptools.setup(
    version=__version__,
    name='secondshot',
    description='Linux-based backup utility',
    long_description=__long_desc__,
    long_description_content_type='text/markdown',
    keywords='backup rsync rsnapshot',
    author='Rich Braun',
    author_email='richb@instantlinux.net',
    url='https://github.com/instantlinux/secondshot',
    entry_points={
      'console_scripts': ['secondshot=secondshot.main:main']
    },
    scripts=['bin/check_rsnap.py', 'bin/cron-secondshot.sh', 'bin/rrsync'],
    packages=setuptools.find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=[
        'alembic>=1.4.3',
        'docopt>=0.6.2',
        'pymysql>=0.10.1',
        'sqlalchemy>=1.3.19'],
    python_requires='>=3.6',
    test_suite='tests.unittests',
    cmdclass={'test': PyTest},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Topic :: System :: Archiving :: Backup',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ]
)



"""actions

Secondshot action logic

created 11-aug-2018 by richb@instantlinux.net

license: lgpl-2.1
"""

import alembic.config
import alembic.script
from alembic.runtime.environment import EnvironmentContext
import binascii
import datetime
import grp
import hashlib
import os
import os.path
import pymysql
import pwd
import socket
from sqlalchemy import create_engine
import sqlalchemy.orm
import stat
import subprocess
import sys
import time

# -*- coding: utf-8 -*-

from secondshot.config import Config
from secondshot.constants import Constants
from secondshot.models import File, Host, Saveset, Volume, metadata, \
    AlembicVersion
from secondshot.syslogger import Syslog


class Actions(object):

    def __init__(self, opts, db_engine=None, db_session=None):
        """Initialize database session"""

        if (db_session):
            self.engine = db_engine
            self.session = db_session
        else:
            self.db_url = Config.get_db_url(opts)
            self.engine = create_engine(self.db_url)
            self.session = sqlalchemy.orm.scoped_session(
                sqlalchemy.orm.sessionmaker(autocommit=False,
                                            bind=self.engine))
        self.set_options(opts)

    def set_options(self, cli_opts):
        """Set configs and define runtime instance variables

        Args:
            cli_opts (dict): command-line as parsed by docopt
        """

        cfg = Config()
        if (cli_opts['backup-host']):
            self.backup_host = cli_opts['backup-host']
        else:
            self.backup_host = socket.gethostname().split('.')[0]
        runtime = cfg.set_opts(cli_opts, self.session, self.backup_host)

        Syslog.logger.debug("Running with options %s" % str(runtime))

        self.filter = runtime['filter'].replace('*', '%')
        self.hosts = runtime['host']
        self.rsnapshot_cfg = cfg.rsnapshot_cfg()
        self.time_fmt = '%Y-%m-%d %H:%M:%S'
        self.volume = runtime['volume']

        if ('snapshot_root' in self.rsnapshot_cfg):
            Config.snapshot_root = self.rsnapshot_cfg[
                'snapshot_root'].rstrip('/')
        self.manifest_file = os.path.join(
            Config.snapshot_root, Constants.SYNC_PATH,
            Config.manifest)
        if ('retain' in self.rsnapshot_cfg):
            self.intervals = self.rsnapshot_cfg['retain']
        elif ('interval' in self.rsnapshot_cfg):
            self.intervals = self.rsnapshot_cfg['interval']
        else:
            self.intervals = Config.sequence

    def calc_sums(self, saveset_id):
        """Calculate checksums for any files that haven't already been stored

        Args:
            saveset_id (int): record ID of saveset
        Returns:
            result (dict):    results summary
        """
        try:
            host = self.session.query(Saveset).filter_by(
                id=saveset_id).one().host.hostname
            location = self.session.query(Saveset).filter_by(
                id=saveset_id).one().location
            saveset = self.session.query(Saveset).filter_by(
                id=saveset_id).one().saveset
        except Exception as ex:
            Syslog.logger.warn('action=calc_sums msg=%s' % str(ex))
            Syslog.logger.traceback(ex)
        if (not host or not location):
            sys.exit('action=calc_sums msg=missing host/location')
        Syslog.logger.info('START action=calc_sums saveset=%s from host=%s '
                           'for location=%s' % (saveset, host, location))
        manifest_file = os.path.join(
            Config.snapshot_root, location, Config.manifest)
        with open(manifest_file, 'r+') as mfile:
            mfile.readline()
            position = mfile.tell()
            (numbytes, count, total) = (0, 0, 0)
            for line in mfile:
                file_id, file_type, size, has_sum = line.strip().split(',')
                total += int(size)
                if has_sum == 'Y' or file_type != 'f' or int(size) == 0:
                    continue
                count += 1
                try:
                    file = self.session.query(File).filter_by(id=file_id).one()
                    filename = os.path.join(
                        Config.snapshot_root, location, file.path,
                        file.filename)
                    file.shasum = self._filehash(filename, Config.hashtype)
                    self.session.add(file)
                    mfile.seek(position)
                    mfile.write('%s,%s,%s,Y\n' % (file_id, file_type, size))
                    numbytes += file.size
                except Exception as ex:
                    Syslog.logger.warn('action=calc_sums id=%d msg=skipped '
                                       'error=%s' % (file.id, str(ex)))
                if (count % 1000 == 0):
                    Syslog.logger.debug('action=calc_sums count=%d bytes=%d' %
                                        (count, numbytes))
                    self.session.commit()
                position = mfile.tell()

        self.session.commit()
        Syslog.logger.info('FINISHED action=calc_sums saveset=%s '
                           'processed=%.3fGB' %
                           (saveset, float(numbytes) / 1e9))
        return {'calc_sums': dict(
            status='ok', saveset=saveset, size=total, processed=numbytes)}

    def inject(self, host, volume, pathname, saveset_id):
        """Inject filesystem metadata for each file in a saveset into manifest

        Args:
            host (str):       host from which to copy files
            volume (str):     saveset's volume name
            pathname (str):   path where current backup is stored
            saveset_id (int): record ID of new saveset
        Returns:
            result (dict):    results summary
        """
        try:
            host_record = self.session.query(Host).filter_by(
                hostname=host).one()
            saveset = self.session.query(Saveset).filter_by(
                id=saveset_id).one()
        except Exception as ex:
            sys.exit('action=inject Invalid host or volume: %s' % str(ex))

        mfile = open(os.path.join(pathname, Config.manifest), 'w')
        mfile.write('file_id,type,file_size,has_checksum\n')

        (count, numbytes, skipped) = (0, 0, 0)
        for dirpath, _, filenames in os.walk(pathname):
            for filename in filenames:
                if filename == Config.manifest:
                    continue
                try:
                    stat = os.lstat(os.path.join(dirpath, filename))
                    _path = pymysql.escape_string(os.path.relpath(
                        dirpath, Config.snapshot_root + '/' +
                        Constants.SYNC_PATH).encode(
                        'utf8', 'surrogateescape').decode('utf8'))
                    _filename = pymysql.escape_string(filename.encode(
                        'utf8', 'surrogateescape').decode('utf8'))
                except OSError as ex:
                    if ex.errno != 2:
                        Syslog.logger.error(
                            'action=inject filename=%s message=%s' %
                            (filename, str(ex)))
                        raise
                    skipped += 1
                    Syslog.logger.debug('action=inject path=%s filename=%s '
                                        'msg=%s' %
                                        (dirpath, filename, str(ex)))
                    continue
                except UnicodeDecodeError as ex:
                    msg = 'action=inject inode=inode=%d dev=%s' % (
                        stat.st_ino, stat.st_dev)
                    try:
                        msg += ' path=%s filename=%s msg=%s' % (
                            dirpath, filename, str(ex))
                    except Exception:
                        pass
                    skipped += 1
                    Syslog.logger.debug(msg)
                    continue
                record = dict(
                    path=_path,
                    filename=_filename,
                    ctime=datetime.datetime.fromtimestamp(
                        stat.st_ctime).strftime(self.time_fmt),
                    gid=stat.st_gid,
                    last_backup=Syslog._now().strftime(
                        '%Y-%m-%d %H:%M:%S'),
                    links=stat.st_nlink,
                    mode=stat.st_mode,
                    mtime=datetime.datetime.fromtimestamp(
                        stat.st_mtime).strftime(self.time_fmt),
                    size=stat.st_size,
                    sparseness=1,
                    type=self._filetype(stat.st_mode),
                    uid=stat.st_uid,
                    host_id=host_record.id)
                try:
                    owner = pwd.getpwuid(stat.st_uid).pw_name
                    group = grp.getgrgid(stat.st_gid).gr_name
                except KeyError:
                    owner = None
                    group = None

                for retry in range(4):
                    try:
                        # Bypass sqlalchemy for ON DUPLICATE KEY UPDATE and
                        # LAST_INSERT_ID functionality
                        if (self.engine.name == 'sqlite'):
                            # sqlite lacks UPSERT capability, but this is
                            # good enough for existing unit-test. TODO: fix
                            #  this so we don't require a 'real' SQL.
                            sql_insert = (
                                u"INSERT OR IGNORE INTO files (%(columns)s)"
                                u" VALUES('%(values)s');" % dict(
                                    columns=','.join(record.keys()),
                                    values="','".join(str(item) for item
                                                      in record.values()),
                                    owner=owner, group=group))
                            sql_id = 'LAST_INSERT_ROWID()'
                        else:
                            sql_insert = (
                                u"INSERT INTO files (%(columns)s)"
                                u" VALUES('%(values)s')"
                                u" ON DUPLICATE KEY UPDATE owner='%(owner)s',"
                                u"grp='%(group)s',id=LAST_INSERT_ID(id),"
                                u"last_backup=NOW();" % dict(
                                    columns=','.join(record.keys()),
                                    values="','".join(str(item) for item
                                                      in record.values()),
                                    owner=owner, group=group))
                            sql_id = 'LAST_INSERT_ID()'
                        self.session.execute(sql_insert)
                        break
                    except sqlalchemy.exc.OperationalError as ex:
                        Syslog.logger.warn('action=inject path=%s filename=%s '
                                           'msg=%s' %
                                           (_path, _filename, str(ex)))
                        if ('Deadlock found' in str(ex)):
                            time.sleep((retry + 1) * 10)
                        else:
                            time.sleep(1)
                    except Exception as ex:
                        Syslog.logger.warn('action=inject path=%s filename=%s '
                                           'msg=%s' %
                                           (_path, _filename, str(ex)))
                        time.sleep(1)
                        raise
                    if (retry == 4):
                        skipped += 1
                file_id = self.session.execute(
                    'SELECT %s' % sql_id).fetchone()[0]
                file = self.session.query(File).filter_by(id=file_id).one()
                has_sha = 'Y' if file.shasum else 'N'
                mfile.write('%d,%s,%d,%s\n' % (
                    file_id, self._filetype(stat.st_mode), stat.st_size,
                    has_sha))
                count += 1
                numbytes += stat.st_size
                if (count % Constants.MAX_INSERT == 0):
                    Syslog.logger.debug('action=inject count=%d' % count)
                    self.session.commit()

        mfile.close()
        self.session.commit()
        saveset.finished = sqlalchemy.func.now()
        saveset.files = count
        saveset.size = numbytes
        self.session.add(saveset)
        self.session.commit()
        Syslog.logger.info('FINISHED action=inject saveset=%s, file_count=%d, '
                           'skipped=%d' % (saveset.saveset, count, skipped))
        return {'inject': dict(
            status='ok', saveset=saveset.saveset, file_count=count,
            skipped=skipped)}

    def rotate(self, interval):
        """Rotate backup entries based on specified interval
        Args:
            interval (str): Values such hourly, daysago as defined by
                            'interval' keyword in rsnapshot conf file

        Returns:
            response (dict): rotation actions taken, if any
        """

        results = []
        if (interval not in self.intervals):
            sys.exit(
                'action=rotate interval=%s must be in %s' %
                (interval, ','.join(sorted(self.intervals.keys()))))
        host_record = self.session.query(Host).filter_by(
            hostname=self.backup_host).one()
        interval_max = int(self.intervals[interval])

        try:
            ret = subprocess.call(['rsnapshot', '-c',
                                   Config.rsnapshot_conf, interval])
        except Exception as ex:
            msg = 'action=rotate subprocess error=%s' % str(ex)
            Syslog.logger.error(msg)
            Syslog.logger.traceback(ex)
            sys.exit(msg)
        if (ret != 0):
            msg = 'action=rotate subprocess returned=%d' % ret
            Syslog.logger.error(msg)
            sys.exit(msg)

        # figure out oldest item in previous interval
        if (interval in ['hourly', 'main', 'short']):
            prev = Constants.SYNC_PATH
        elif (interval in Config.sequence):
            prev = Config.sequence[Config.sequence.index(interval) - 1]
            prev += ".%d" % (int(self.intervals[prev]) - 1)
        else:
            sys.exit('action=rotate interval=%s unrecognized' % interval)

        # delete savesets that match <interval>.<interval_max - 1>
        count = self.session.query(Saveset).filter_by(
            location='%s.%d' % (interval, interval_max - 1),
            backup_host_id=host_record.id).delete()
        if (count > 0):
            results.append(dict(
                action='delete',
                host=self.backup_host,
                location='%s.%d' % (interval, interval_max - 1),
                savesets=count))
            Syslog.logger.info(
                'action=rotate host=%s location=%s.%d savesets=%d removed' %
                (self.backup_host, interval, interval_max - 1, count))

        # move all savesets location <interval>.<n> => <n+1>
        if (self.engine.name == 'sqlite'):
            self.session.execute(
                "UPDATE savesets SET location='%(interval)s'||'.'||"
                "(SUBSTR(location,INSTR(location,'.')+1)+1) WHERE location "
                "LIKE '%(interval)s%%' AND backup_host_id=%(host)d" % {
                    'interval': interval, 'host': host_record.id})
        else:
            self.session.execute(
                "UPDATE savesets SET location=CONCAT('%(interval)s','.',"
                "SUBSTR(location,INSTR(location,'.')+1)+1) WHERE location "
                "LIKE '%(interval)s%%' AND backup_host_id=%(host)d" % {
                    'interval': interval, 'host': host_record.id})

        # move saveset location=<previous int> to <interval>.0
        count = self.session.query(Saveset).filter(
            Saveset.location == prev,
            Saveset.backup_host_id == host_record.id,
            Saveset.finished is not None).update({
                Saveset.location: '%s.0' % interval})
        if (count > 0):
            results.append(dict(
                host=self.backup_host,
                savesets=count,
                location='%s.0' % interval,
                prev=prev))
            Syslog.logger.info('action=rotate host=%s savesets=%d '
                               'location=%s.0 prev=%s' %
                               (self.backup_host, count, interval, prev))
        self.session.commit()
        return {'rotate': dict(status='ok' if results else 'error',
                               actions=results)}

    def start(self, hosts, volume):
        """Start a backup for each of the specified hosts; if
        successful, also calculate sha checksums for any missing
        entries and re-read each stored file to verify

        Args:
            hosts (list):      hosts to back up
            volume (str):      volume path of destination
        Returns:
            dict: operations performed
        """
        if (len(hosts) == 0):
            sys.exit('action=start must specify at least one --host')
        results = []
        status = 'ok'
        for host in hosts:
            try:
                new_saveset = self.new_saveset(host, volume)
                saveset_id = new_saveset['id']
            except sqlalchemy.exc.IntegrityError as ex:
                Syslog.logger.error('action=start database error=%s'
                                    % str(ex))
                status = 'error'
                continue
            try:
                ret = subprocess.call(['rsnapshot', '-c',
                                       Config.rsnapshot_conf, 'sync', host])
            except Exception as ex:
                Syslog.logger.error('action=start subprocess error=%s'
                                    % str(ex))
                status = 'error'
                continue
            if (ret != 0):
                Syslog.logger.error('action=start rsnapshot process error=%d'
                                    % ret)
                status = 'error'
                continue
            try:
                results.append(
                    self.inject(host, volume, '%s/%s' %
                                (Config.snapshot_root, Constants.SYNC_PATH),
                                saveset_id))
            except Exception as ex:
                Syslog.logger.error('action=start inject error=%s' % str(ex))
                Syslog.logger.traceback(ex)
                status = 'error'
                continue
            results.append(self.calc_sums(saveset_id))
            if (Config.autoverify):
                try:
                    result = self.verify([new_saveset['saveset']])
                    results.append(result)
                    if (result['verify']['status'] != 'ok'):
                        status = 'error'
                except RuntimeError:
                    status = 'error'

        return {'start': dict(status=status, results=results)}

    def new_saveset(self, host, volume):
        """Set up new saveset entry in database
        Args:
            host (str):   host from which to copy files
            volume (str): volume path of destination

        Returns:
            dict: id and name of new record in saveset table
        """

        vol = self.session.query(Volume).filter_by(volume=volume).one()
        host_record = self.session.query(Host).filter_by(
            hostname=host).one()
        backup_host_record = self.session.query(Host).filter_by(
            hostname=self.backup_host).one()
        if (not host or not vol):
            sys.exit('Invalid host or volume')
        saveset = Saveset(
            saveset='%(host)s-%(volume)s-%(date)s' % dict(
                host=host,
                volume=volume,
                date=Syslog._now().strftime('%Y%m%d-%H')),
            location=Constants.SYNC_PATH,
            host=host_record,
            backup_host=backup_host_record
        )
        try:
            self.session.add(saveset)
            self.session.commit()
        except sqlalchemy.exc.IntegrityError as ex:
            if ('Duplicate entry' in str(ex)):
                sys.exit('ERROR: duplicate saveset=%s' % saveset.saveset)
        Syslog.logger.info('START saveset=%s' % saveset.saveset)
        return dict(id=saveset.id, saveset=saveset.saveset)

    def verify(self, savesets):
        """Read each file in specified savesets to verify against stored
        checksums

        Parameters:
            savesets (list): saveset names
        Returns:
            result (dict): summary of files checked
                  status = ok if no errors
                  [count = files examined
                   skipped = files that couldn't be read (e.g. permissions)
                   missing = files for which the DB has no stored checksum
                   errors = files with content that does not match checksum]
        Raises:
            RuntimeError: if saveset is missing
        """

        results = []
        for saveset in savesets:
            try:
                record = self.session.query(Saveset).filter_by(
                        saveset=saveset).one()
            except sqlalchemy.orm.exc.NoResultFound:
                raise RuntimeError('VERIFY saveset=%s not found' % saveset)

            manifest_file = os.path.join(
                Config.snapshot_root, record.location, Config.manifest)
            mfile = open(manifest_file, 'r')
            mfile.readline()
            count, errors, missing, skipped = (0, 0, 0, 0)
            for line in mfile:
                file_id, file_type, size, has_sum = line.strip().split(',')
                if has_sum != 'Y' or file_type != 'f' or int(size) == 0:
                    continue
                count += 1
                file = self.session.query(File).filter_by(id=file_id).one()
                if (Config.hashtype != self._hashtype(file.shasum)):
                    Config.hashtype = self._hashtype(file.shasum)
                    Syslog.logger.info('action=verify hashtype=%s'
                                       % Config.hashtype)
                try:
                    filename = os.path.join(
                        Config.snapshot_root, record.location, file.path,
                        file.filename)
                    sha = self._filehash(filename, Config.hashtype)
                    if (sha != file.shasum):
                        Syslog.logger.warn(
                            'BAD CHECKSUM: action=verify file=%s/%s '
                            'expected=%s actual=%s' %
                            (file.path, file.filename,
                             binascii.hexlify(file.shasum),
                             binascii.hexlify(sha)))
                        errors += 1
                except Exception as ex:
                    Syslog.logger.debug('sha(%s): %s' % (
                        file.filename, str(ex)))
                    skipped += 1
                if (count % 1000 == 0):
                    Syslog.logger.debug('action=verify count=%d skipped=%d '
                                        'errors=%d' % (count, skipped, errors))
            mfile.close()
            msg = ('VERIFY: saveset=%s count=%d errors=%d missing=%d '
                   'skipped=%d' % (saveset, count, errors, missing, skipped))
            if (errors):
                Syslog.logger.error(msg)
            else:
                Syslog.logger.info(msg)
            results.append(dict(
                saveset=saveset, count=count, errors=errors,
                missing=missing, skipped=skipped))

        return {'verify': dict(
            status='ok' if errors == 0 else 'error',
            results=results)}

    def schema_update(self):
        """Examines the Alembic schema version and performs database
        migration if needed

        Returns:
            result (dict): summary of changes made
                  status = ok if no errors
        """

        results = []
        errors = 0
        try:
            version = self.session.query(AlembicVersion).one().version_num
        except (sqlalchemy.orm.exc.NoResultFound,
                sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            Syslog.logger.warn('DB schema does not yet exist: %s' % str(ex))
            version = None
        cfg = alembic.config.Config()
        cfg.set_main_option('script_location', os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'alembic'))
        cfg.set_main_option('url', str(self.engine.url))
        script = alembic.script.ScriptDirectory.from_config(cfg)
        env = EnvironmentContext(cfg, script)
        if (version == script.get_heads()[0]):
            Syslog.logger.info('action=schema-update version=%s is current, '
                               'skipping' % version)
            results.append(dict(name=version, action='skipped'))
        else:
            def _do_upgrade(revision, context):
                return script._upgrade_revs(script.get_heads(), revision)

            conn = self.engine.connect()
            env.configure(connection=conn, target_metadata=metadata,
                          fn=_do_upgrade)
            with env.begin_transaction():
                env.run_migrations()
            results.append(dict(name=script.get_heads()[0], action='migrated'))
            Syslog.logger.info('action=schema-update finished migration, '
                               'version=%s' % script.get_heads()[0])
            if (version is None):
                # Seed a new db with host and volume records

                record = Host(hostname=self.backup_host)
                self.session.add(record)
                self.session.flush()
                self.session.add(Volume(
                    volume=Constants.DEFAULT_VOLUME,
                    path=Constants.SNAPSHOT_ROOT,
                    host_id=record.id))
                self.session.commit()
        return {
            'status': 'ok' if errors == 0 else 'error',
            'schema-update': results}

    def list_hosts(self):
        """List hosts"""
        items = self.session.query(Host).filter(
            Host.hostname.like(self.filter)).order_by('hostname')
        return {'hosts': [dict(
                    name=item.hostname,
                    created=item.created.strftime(self.time_fmt)
                    ) for item in items]}

    def list_savesets(self):
        """List savesets"""
        items = self.session.query(Saveset).filter(
            Saveset.saveset.like(self.filter)).order_by('saveset')
        return {'savesets': [dict(
                    name=item.saveset, location=item.location,
                    created=item.created.strftime(self.time_fmt),
                    host=item.host.hostname,
                    backup_host=item.backup_host.hostname,
                    files=item.files,
                    size=item.size,
                    finished=item.finished.strftime(self.time_fmt) if
                             item.finished else None) for item in items]}

    def list_volumes(self):
        """List volumes"""
        items = self.session.query(Volume).filter(
            Volume.volume.like(self.filter)).order_by('volume')
        return {'volumes': [dict(
                    name=item.volume, path=item.path, size=item.size,
                    created=item.created.strftime(self.time_fmt),
                    host=item.host.hostname) for item in items]}

    @staticmethod
    def _filehash(file, hashtype):
        """Read a file and return its hash

        Args:
            file (str): name of file
            hashtype (str): type of hash
        Returns:
            bytes:  binary digest (as 16, 32 or 64 bytes)
        Raises:
            OS exceptions
        """
        with open(file, 'rb') as f:
            if (hashtype == 'md5'):
                return hashlib.md5(f.read()).digest()
            elif (hashtype == 'sha256'):
                return hashlib.sha256(f.read()).digest()
            elif (hashtype == 'sha512'):
                return hashlib.sha512(f.read()).digest()

    @staticmethod
    def _hashtype(shasum):
        """Identify hashtype of a shasum value

        Args:
            shasum (str): binary string
        Returns:
            bytes:  hash type md5, sha256 or sha512
        Raises:
            RuntimeError on failure
        """
        if (len(shasum) == 16):
            return 'md5'
        elif (len(shasum) == 32):
            return 'sha256'
        elif (len(shasum) == 64):
            return 'sha512'
        else:
            msg = ('Hash type unrecognized')
            Syslog.logger.error(msg)
            raise RuntimeError(msg)

    @staticmethod
    def _filetype(mode):
        """Determine file type given the stat mode bits

        Args:
           mode (int): mode bits returned by os.stat()
        Returns:
           char:  character-special, directory, file, link (symbolic),
                  pipe, socket
        Raises:
            RuntimeError: if unexpected status
        """

        if (stat.S_ISCHR(mode)):
            return 'c'
        elif (stat.S_ISDIR(mode)):
            return 'd'
        elif (stat.S_ISREG(mode)):
            return 'f'
        elif (stat.S_ISLNK(mode)):
            return 'l'
        elif (stat.S_ISSOCK(mode)):
            return 's'
        elif (stat.S_ISFIFO(mode)):
            return 'p'
        else:
            raise RuntimeError('Unexpected file mode %d' % mode)



"""add config table

Revision ID: 23296991a8e5
Revises: ab8ec674d281
Create Date: 2018-08-15 11:53:00.857193

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '23296991a8e5'
down_revision = 'ab8ec674d281'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'config',
        sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column('host_id', sa.INTEGER(), nullable=False),
        sa.Column('keyword', sa.String(length=32), nullable=False),
        sa.Column('value', sa.String(length=1023), nullable=True),
        sa.Column('created', sa.TIMESTAMP(), server_default=sa.func.now(),
                  nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id')
    )


def downgrade():
    op.drop_table('config')



"""syslogger

Syslog functions

created 11-aug-2018 by richb@instantlinux.net

license: lgpl-2.1
"""

import datetime
import logging
import logging.handlers
import os.path
import sys

logfile = None
logger = logging.getLogger()
syslog_enable = True


class Syslog(object):

    def __init__(self, opts):
        global logfile, logger, syslog_enable

        syslog_enable = True
        self.prog = 'secondshot'
        self.log_facility = 'local1'
        if (opts['verbose'] or opts['log-level'].lower() == 'debug'):
            self.log_level = logging.DEBUG
        elif (opts['log-level'].lower() == 'info'):
            self.log_level = logging.INFO
        elif (opts['log-level'].lower() == 'warn'):
            self.log_level = logging.WARN
        elif (opts['log-level'].lower() == 'none'):
            self.log_level = logging.INFO
            syslog_enable = False
        else:
            sys.stderr.write(
                'Unrecognized --log-level=%s\n' % opts['log-level'])
            self.log_level = logging.INFO
        if (syslog_enable):
            logger.setLevel(self.log_level)
            if (os.path.exists('/dev/log')):
                handler = logging.handlers.SysLogHandler(
                    address='/dev/log', facility=self.log_facility)
                logger.addHandler(handler)
            else:
                syslog_enable = False
        logfile = opts['logfile']

    def debug(self, msg):
        global logfile, logger, syslog_enable
        if (self.log_level == logging.DEBUG):
            sys.stderr.write('DEBUG: %s\n' % msg)
            if (syslog_enable):
                logger.debug('%s %s' % (self.prog, msg))
            with open(logfile, 'a') as f:
                f.write(self._date_prefix('D', msg))

    def error(self, msg):
        global logfile, logger, syslog_enable
        sys.stderr.write('ERROR: %s\n' % msg)
        if (syslog_enable):
            logger.error('%s %s' % (self.prog, msg))
        with open(logfile, 'a') as f:
            f.write(self._date_prefix('E', msg))

    def info(self, msg):
        global logfile, logger, syslog_enable
        if (self.log_level == logging.DEBUG):
            sys.stderr.write('INFO: %s\n' % msg)
        if (syslog_enable):
            logger.info('%s %s' % (self.prog, msg))
        with open(logfile, 'a') as f:
            f.write(self._date_prefix('I', msg))

    def warn(self, msg):
        global logfile, logger, syslog_enable
        sys.stderr.write('WARN: %s\n' % msg)
        if (syslog_enable):
            logger.warning('%s %s' % (self.prog, msg))
        with open(logfile, 'a') as f:
            f.write(self._date_prefix('W', msg))

    def _date_prefix(self, severity, msg):
        return '[%s] %s %s\n' % (
            self._now().strftime('%d/%b/%Y-%H:%M:%S'),
            severity, msg)

    @staticmethod
    def _now():
        """Convenience mathod wraps datetime.now() for unit-test mocks

        Returns:
            obj (datetime): current time
        """
        return datetime.datetime.now()
